/***********************************************************
Copyright 1991 by Digital Equipment Corporation, Maynard, Massachusetts,
the Massachusetts Institute of Technology, Cambridge, Massachusetts,
and Olivetti Research Limited, Cambridge, England.

                        All Rights Reserved

Permission to use, copy, modify, and distribute this software and its 
documentation for any purpose and without fee is hereby granted, 
provided that the above copyright notice appear in all copies and that
both that copyright notice and this permission notice appear in 
supporting documentation, and that the names of Digital, Olivetti or MIT
not be used in advertising or publicity pertaining to distribution of the
software without specific, written prior permission.  

DIGITAL AND OLIVETTI DISCLAIM ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO EVENT
SHALL THEY BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR
ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN
AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR
IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

******************************************************************/

#ifndef _SYNC_H_
#define _SYNC_H_

#define SYNCNAME "ORL-SYNC"

#define SYNC_MAJOR_VERSION	2
#define SYNC_MINOR_VERSION	0

#define X_SyncInfo			0
#define X_SyncCreateCounter		1
#define X_SyncDestroyCounter		2
#define X_SyncQueryCounter              3
#define X_SyncAwait			4
#define X_SyncAdvance			5
#define X_SyncCreateAlarm               6
#define X_SyncDestroyAlarm		7
#define X_SyncQueryAlarm		8
#define X_SyncChangeAlarm		9

#define AdvanceNotify			0
#define AdvanceNotifyMask		(1L << AdvanceNotify)

#define AlarmNotify			1
#define AlarmNotifyMask 		(1L << AlarmNotify)
#define SyncNumberEvents		2

#define BadCounter			0
#define BadAlarm			1
#define SyncNumberErrors		2

/*
 * Flags for Alarm Attributes
 */
#define CACounter			(1L<<0)
#define CAValueType			(1L<<1)
#define CAValue				(1L<<2)
#define CAIncrement			(1L<<3)
#define CALimitType			(1L<<4)
#define CALimit				(1L<<5)
#define CAEvents			(1L<<6)

/*
 * Constants for the value_type argument of various requests
 */
#define SyncAbsolute		1
#define SyncRelative		2

/*
 * Alarm state constants
 */
#define AlarmActive			1
#define AlarmExhausted			2
#define AlarmDestroyed			3

#if LongIs64Bits
/*
 * XXX: This code is untested, for obvious reasons!
 */
typedef long CARD64;

#define cvtLongToC64(c64, i)	(c64 = (i))
#define cvtC64ToLong(c64)	(c64)
#define cvtINT32sToC64(c64, l, h)((c64) = ((h)<<32) || (l))
#define C64GreaterThan(a, b)	((a)>(b))
#define C64LessThan(a, b)	((a)<(b))
#define C64GreaterOrEqual(a, b)	((a)>=(b))
#define C64LessOrEqual(a, b)	((a)<=(b))
#define C64Equal(a, b)		((a)==(b))
#define C64IsZero(a)		((a)==0)
#define C64Low32(c64)		((c64)&0xFFFFFFFF)
#define C64High32(c64)		((c64>>32)&0xFFFFFFFF)
#define C64Add(res,a,b)		((res) = (a) + (b))
#define C64Subtract(res,a,b)	((res) = (a) - (b))
#define MAXCARD64(v)		((v)=0xFFFFFFFFFFFFFFFF);

#else /* LongIs64Bits */

typedef struct CARD64Struct {
    unsigned long lo;
    unsigned long hi;
} CARD64;

#define cvtLongToC64(c64, i)	((c64).hi=0,(c64).lo=(i))
#define cvtC64ToLong(c64)	((c64).lo)
#define cvtINT32sToC64(c64, l, h) ((c64).lo = (l), (c64).hi = (h))
#define C64GreaterThan(a, b)\
    ((a).hi>(b).hi || ((a).hi==(b).hi && (a).lo>(b).lo))
#define C64LessThan(a, b)\
    ((a).hi<(b).hi || ((a).hi==(b).hi && (a).lo<(b).lo))
#define C64GreaterOrEqual(a, b)\
    ((a).hi>(b).hi || ((a).hi==(b).hi && (a).lo>=(b).lo))
#define C64LessOrEqual(a, b)\
    ((a).hi<(b).hi || ((a).hi==(b).hi && (a).lo<=(b).lo))
#define C64Equal(a, b)		((a).lo==(b).lo && (a).hi==(b).hi)
#define C64IsZero(a)		((a).lo==0 && (a).hi==0)
#define C64Low32(c64)		((c64).lo)
#define C64High32(c64)		((c64).hi)
#define C64Add(res,a,b) {\
	int t = (a).lo;\
	((res).lo = (a).lo + (b).lo);\
	((res).hi = (a).hi + (b).hi);\
	if (t>(res).lo) (res).hi++;\
     }
#define C64Subtract(res,a,b) {\
	int t = (a).lo;\
	((res).lo = (a).lo - (b).lo);\
	((res).hi = (a).hi - (b).hi);\
	if (t>(res).lo) (res).hi--;\
     }
#define MAXCARD64(v)		((v).lo=0xFFFFFFFF,(v).hi=0xFFFFFFFF);

#endif /* LongIs64Bits */

#ifndef _SYNC_SERVER

typedef long Counter;
typedef long Alarm;

typedef struct _SystemCounterRec {
    char *name;
    Counter counter;
    CARD64 resolution;
} SystemCounterRec, *SystemCounterList;

typedef struct {
    int type;
    unsigned long serial;
    Bool send_event;
    Display *display;
    Counter counter;
    CARD64 wait_value;
    CARD64 counter_value;
    Time time;
} XAdvanceNotifyEvent;

typedef struct {
    int type;
    unsigned long serial;
    Bool send_event;
    Display *display;
    Alarm alarm;
    CARD64 counter_value;
    CARD64 alarm_value;
#ifdef XXX
    CARD64 alarm_increment;
    CARD64 alarm_limit;
#endif /* XXX */
    int state;
    Time time;
} XAlarmNotifyEvent;

typedef struct {
    Counter counter;
    int value_type;
    CARD64 value;
    CARD64 increment;
    int limit_type;
    CARD64 limit;
    int events;
    int state;
} XSyncAlarmAttributes;

typedef struct {
    Counter counter;
    int wait_type;
    CARD64 wait_value;
    CARD64 event_threshold;
} XWaitCondition, *XWaitConditionList;

extern Bool XSyncQueryExtension(
#if NeedFunctionPrototypes
    Display*		/* dpy */,
    unsigned long*	/* event_return */,
    unsigned long*	/* error_return */
#endif
);

extern Status XSyncInfo(
#if NeedFunctionPrototypes
    Display*		/* dpy */,
    int*		/* major_return */,
    int*		/* minor_return */,
    SystemCounterList*	/* list_return */,
    unsigned long*     	/* n_counters_return */
#endif
);

extern Counter XSyncCreateCounter(
#if NeedFunctionPrototypes
    Display*		/* dpy */,
    CARD64	        /* initial_value */
#endif
);

extern Counter XSyncCreateCounter32(
#if NeedFunctionPrototypes
    Display*		/* dpy */,
    unsigned long       /* initial_value */
#endif
);

extern Status XSyncDestroyCounter(
#if NeedFunctionPrototypes
    Display*		/* dpy */,
    Counter		/* counter */
#endif
);

extern Status XSyncQueryCounter(
#if NeedFunctionPrototypes
    Display*		/* dpy */,
    Counter		/* counter */,
    CARD64*	        /* value_return */
#endif
);

extern Status XSyncQueryCounter32(
#if NeedFunctionPrototypes
    Display*		/* dpy */,
    Counter		/* counter */,
    unsigned long*	/* value_return */
#endif
);

extern Status XSyncAwait(
#if NeedFunctionPrototypes
    Display *           /* dpy */,
    XWaitConditionList  /* wait_list */,
    unsigned long       /* n_conditions */
#endif
);

extern Status XSyncAdvance(
#if NeedFunctionPrototypes
    Display*		/* dpy */,
    Counter		/* counter */,
    CARD64	        /* amount */
#endif
);

extern Status XSyncAdvance32(
#if NeedFunctionPrototypes
    Display*		/* dpy */,
    Counter		/* counter */,
    unsigned long       /* amount */
#endif
);

extern Alarm XSyncCreateAlarm(
#if NeedFunctionPrototypes
    Display*		/* dpy */,
    unsigned long       /* values_mask */,
    XSyncAlarmAttributes* /* values */
#endif
);

extern Status XSyncDestroyAlarm(
#if NeedFunctionPrototypes
    Display*		/* dpy */,
    Alarm		/* alarm*/
#endif
);

extern Status XSyncQueryAlarm(
#if NeedFunctionPrototypes
    Display*		/* dpy */,
    Alarm		/* alarm */,
    XSyncAlarmAttributes* /* values_return */
#endif
);

extern Status XSyncChangeAlarm(
#if NeedFunctionPrototypes
    Display*		/* dpy */,
    Alarm		/* alarm */,
    unsigned long	/* values_mask */,
    XSyncAlarmAttributes* /* values */
#endif
);

#endif /* _SYNC_SERVER */

#endif /* _SYNC_H_ */
