/***********************************************************
Copyright 1991 by Digital Equipment Corporation, Maynard, Massachusetts,
the Massachusetts Institute of Technology, Cambridge, Massachusetts,
and Olivetti Research Limited, Cambridge, England.

                        All Rights Reserved

Permission to use, copy, modify, and distribute this software and its 
documentation for any purpose and without fee is hereby granted, 
provided that the above copyright notice appear in all copies and that
both that copyright notice and this permission notice appear in 
supporting documentation, and that the names of Digital, Olivetti or MIT
not be used in advertising or publicity pertaining to distribution of the
software without specific, written prior permission.  

DIGITAL AND OLIVETTI DISCLAIM ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO EVENT
SHALL THEY BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR
ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN
AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR
IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

******************************************************************/

#ifdef SYNC

#define CHECK

#define NEED_REPLIES
#define NEED_EVENTS
#include <stdio.h>
#include "X.h"
#include "Xproto.h"
#include "misc.h"
#include "os.h"
#include "extnsionst.h"
#include "dixstruct.h"
#include "resource.h"
#include "opaque.h"
#define _SYNC_SERVER
#include "syncstr.h"

/*
 * External routines that should be defined in header files
 */

extern ExtensionEntry *AddExtension(/* LOTSOFARGS */);

/*
 * Local types
 */
typedef struct _ClientWaitItem {
    CARD64  		threshold;	/* event threshold */
    struct _WaitList	*wait_ptr;
} ClientWaitItem;

typedef struct _AlarmWaitItem {
    struct _SyncAlarm	*alarm;
} AlarmWaitItem;

typedef struct _WaitItem {
    struct _SyncCounter	*counter;	/* the counter for this item */
    CARD64  		value;		/* the value we are waiting for */
    int			value_type;	/* Absolute or Relative */
    struct _WaitItem	*next;
    struct _WaitItem	*prev;
    int			is_alarm;
    union {
	AlarmWaitItem	a_wait;
	ClientWaitItem	c_wait;
    } u;
} WaitItem;

typedef struct _WaitList {
    struct _AwaitClient	*aclient;
    int		n_items;
    WaitItem	items[1];	/* Must be last */
} WaitList;

typedef struct _AwaitClient {
    XID		id;
    ClientPtr	client;
    WaitList	*wait;
    struct _AwaitClient	*next;
} AwaitClient;

typedef struct _AlarmClient {
    XID		id;
    ClientPtr	client;
    struct _AlarmClient	*next;
} AlarmClient;

typedef struct _SyncAlarm {
    XID 	id;
    ClientPtr	client;
    AlarmClient	*clients;
    int		state;
    CARD64	increment;
    int		limit_type;
    CARD64	limit;
    WaitItem	wait;
} SyncAlarm;

typedef struct _SyncCounter {
    XID		id;		/* resource ID */
    ClientPtr	owner;		/* Owning client. 0 for system counters */
    void	(*Query)(
#if NeedFunctionPrototypes
			 struct _SyncCounter  *,
			 CARD64 *
#endif
			 );
    void	(*NewWait)(
#if NeedFunctionPrototypes
			   struct _SyncCounter  *,
			   CARD64,
			   int,
			   CARD64 *
#endif
			   );
    void	(*NoWait)(
#if NeedFunctionPrototypes
			  struct _SyncCounter  *,
#endif
			  );
    WaitItem	wait;		/* current value + sorted list of clients */
} SyncCounter;

typedef struct _SysCounter {
    char	*name;
    SyncCounter	*counter;
    CARD64	resolution;
} SysCounter;

/*
 * Local Global Variables
 */
static int SyncReqCode;
static int SyncEventBase;
static int SyncErrorBase;
static RESTYPE RTCounter = 0;
static RESTYPE RTAwaitClient;
static RESTYPE RTAlarm;
static RESTYPE RTAlarmClient;

static void FreeAllAlarms();
static void TriggerAwait();

static SysCounter *SysCounterList = NULL;
static int NumSystemCounters = 0;
static AwaitClient *AwaitClientList = NULL;

#ifdef CHECK
static void ShowWaitItem(pWait)
WaitItem *pWait;
{
    fprintf(stderr, "%8.8X <= %8.8X => %8.8X value %X%8.8X %s",
	    pWait->prev, pWait, pWait->next,
	    C64High32(pWait->value),
	    C64Low32(pWait->value),
	    pWait->value_type==SyncAbsolute? "ABS":"REL");
    if (pWait->is_alarm) {
	fprintf(stderr, " Alarm %X\n", pWait->u.a_wait.alarm);
    } else {
	fprintf(stderr, " Client\n");
    }
}

static void ShowCounter(pCounter)
SyncCounter *pCounter;
{
    WaitItem *pWait = &pCounter->wait;

    fprintf(stderr, "Counter: id %X owner %X\n");
    while (pWait != 0) {
	ShowWaitItem(pWait);
	pWait = pWait->next;
    }
    fprintf(stderr, "\n");
}

static int CheckCounter(pCounter, where)
SyncCounter *pCounter;
char *where;
{
    WaitItem *pWait = &pCounter->wait;
    WaitItem *pNext;

    while ((pNext = pWait->next) != 0) {
	if (pWait->counter != pCounter) {
	    fprintf(stderr, "%s: Wrong counter value in WaitItem\n", where);
	    ShowCounter(pCounter);
	    return 0;
	}
	if (C64LessThan(pNext->value,pWait->value) && 
	    pWait != &pCounter->wait) {
	    fprintf(stderr, "%s: Badly ordered counter:\n", where);
	    ShowCounter(pCounter);
	    return 0;
	}
	if (pNext->prev != pWait) {
	    fprintf(stderr, "%s: Badly linked counter:\n", where);
	    ShowCounter(pCounter);
	    return 0;
	}
	pWait = pNext;
    }

    return 1;
}
#endif

/*
*** AddWaitItem
*** Add a new wait item to a counter
 */
static void AddWaitItem(pCounter, pItem)
SyncCounter *pCounter;
WaitItem *pItem;
{
    WaitItem *pTmp;

    if (pCounter==0) {
	return;
    }
/*
 * Inform each counter of this wait and calculate values
 * for relative waits.
 */
    (*pCounter->NewWait)((pointer)pCounter, pItem->value,
			 pItem->value_type,
			 &pCounter->wait.value);
    if (pItem->value_type==SyncRelative) {
	C64Add(pItem->value, pCounter->wait.value, pItem->value);
	pItem->value_type = SyncAbsolute;
    }
/*
 * Now link the item on the list
 */
#ifdef CHECK
    CheckCounter(pCounter,"pre-AddWaitItem");
#endif
    pTmp = &pCounter->wait;
    while (pTmp->next != 0 &&
	   C64GreaterThan(pItem->value, pTmp->next->value)) {
	pTmp = pTmp->next;
    }

    pItem->next = pTmp->next;
    pItem->prev = pTmp;
    pTmp->next = pItem;
    if (pItem->next) {
	pItem->next->prev = pItem;
    }

#ifdef CHECK
    if (!CheckCounter(pCounter,"AddWaitItem")) {
	ShowWaitItem(pItem);
	fprintf(stderr, "... has just been added\n");
    }
#endif
}

/*
*** RemoveWaitItem
*** Remove a wait item from a counter
 */
static void RemoveWaitItem(pCounter, pItem)
SyncCounter *pCounter;
WaitItem *pItem;
{
/*
 * If this item is not on a list, there is nothing to do
 */
    if (pCounter == 0 || pItem->prev == 0) {
	return;
    }
/*
 * If this item is first on the list, tell the counter that it has gone
 */
    if (pItem == pCounter->wait.next) {
	if (pItem->next) {
	    (*pCounter->NewWait)((pointer)pCounter, pItem->next->value,
				 SyncAbsolute, &pCounter->wait.value);
	} else {
	    (*pCounter->NoWait)((pointer)pCounter);
	}
    }
/*
 * Remove this from the list
 */
    pItem->prev->next = pItem->next;
    if (pItem->next) {
	pItem->next->prev = pItem->prev;
    }
    pItem->next = pItem->prev = NULL;
#ifdef CHECK
    if (!CheckCounter(pCounter, "RemoveWaitItem")) {
	ShowWaitItem(pItem);
	fprintf(stderr, "... has just been removed\n");
    }
#endif
}

/*
*** ReorderWaitItem
*** Move a wait item around in the queue, if necessary
 */
static void ReorderWaitItem(pCounter, pItem)
SyncCounter *pCounter;
WaitItem *pItem;
{
    int was_first = pCounter->wait.next == pItem;

/*
 * Move item down the queue
 */
    while (pItem->next != NULL &&
	   C64GreaterThan(pItem->value, pItem->next->value)) {
	WaitItem *pPrev = pItem->prev;
	WaitItem *pNext = pItem->next;

	pPrev->next = pNext;
	pItem->prev = pNext;
	pItem->next = pNext->next;
	pNext->prev = pPrev;
	pNext->next = pItem;
    }

/*
 * If we were head or are still head, tell the counter about the change
 * in wait time.
 */
    if ((was_first || pCounter->wait.next == pItem)) {
	(*pCounter->NewWait)((pointer)pCounter, pCounter->wait.next->value,
			     SyncAbsolute, &pCounter->wait.value);
    }
#ifdef CHECK
    if (!CheckCounter(pCounter,"ReorderWaitItem")) {
	ShowWaitItem(pItem);
	fprintf(stderr, "... has just been re-ordered\n");
    }
#endif
}


/*
*** TriggerAlarm
 */
static void TriggerAlarm(pAlarm)
SyncAlarm *pAlarm;
{
    AlarmClient *pClient;
    SyncCounter *pCounter = pAlarm->wait.counter;
    CARD64 a_value;

    a_value = pAlarm->wait.value;

/*
 * Advance the alarm time and check to see what the new state is
 */
    if (pCounter == 0) {
	pAlarm->state = AlarmExhausted;
    } else if (C64IsZero(pAlarm->increment) ||
	       C64LessOrEqual(pAlarm->limit, pCounter->wait.value)) {
	pAlarm->state = AlarmExhausted;
	RemoveWaitItem(pCounter, &pAlarm->wait);
    } else if (C64LessThan(pCounter->wait.value, pAlarm->wait.value)) {
	return;
    } else {
	while (C64LessOrEqual(pAlarm->wait.value, pCounter->wait.value) ) {
	    if (C64LessOrEqual(pAlarm->limit, pCounter->wait.value)) {
		pAlarm->state = AlarmExhausted;
		RemoveWaitItem(pCounter, &pAlarm->wait);
		break;
	    }
	    C64Add(pAlarm->wait.value, pAlarm->wait.value, pAlarm->increment);
	    ReorderWaitItem(pCounter, &pAlarm->wait);
	}
    }

/*
 * Send an event to all the interested clients
 */
    UpdateCurrentTimeIf();

    pClient = pAlarm->clients;
    while (pClient) {
	ClientPtr client = pClient->client;

	if (client && !client->clientGone) {
	    xAlarmNotifyEvent e;
	    e.type = SyncEventBase + AlarmNotify;
	    e.sequenceNumber = client->sequence;
	    e.alarm = pAlarm->id;
	    if (pCounter) {
		e.counter_value_lo = C64Low32(pCounter->wait.value);
		e.counter_value_hi = C64High32(pCounter->wait.value);
	    } else {
		e.counter_value_lo = 0;
		e.counter_value_hi = 0;
	    }
	    e.alarm_value_lo = C64Low32(a_value);
	    e.alarm_value_hi = C64High32(a_value);
	    e.state = pAlarm->state;
	    e.time = currentTime.milliseconds;
	    WriteEventsToClient(client, 1, (xEvent *)&e);
	}
	pClient = pClient->next;
    }
}

/*
*** Wake up all the clients who's value exceeds the counter value
 */
static void SyncReleaseClients(pCounter)
SyncCounter *pCounter;
{
/*
 * The client list is sorted, so just keep taking the head off the list
 */
    while (pCounter->wait.next &&
	   C64GreaterOrEqual(pCounter->wait.value, pCounter->wait.next->value)) {
	WaitItem *pItem = pCounter->wait.next;
/*
 * Take the action appropriate for Alarms/Clients
 */
	if (pItem->is_alarm) {
	    TriggerAlarm(pItem->u.a_wait.alarm);
	} else {
	    TriggerAwait(pItem->u.c_wait.wait_ptr);
	}
    }
}

/*
*** Wake up all the clients for this counter
 */
static void ReleaseAllClients(pCounter)
SyncCounter *pCounter;
{
    CARD64 c_value;
/*
 * For now, we just set the counter value to maximum.
 * This will mean that ALL clients are released but it may be expensive
 * for Alarms
 */
    MAXCARD64(c_value);
    pCounter->wait.value = c_value;

    SyncReleaseClients(pCounter);
}

/*
*** Cleanup after the destruction of a Counter
 */
/* ARGSUSED */
static void FreeCounter(env, id)
pointer env;
XID id;
{
    SyncCounter *pCounter = (SyncCounter *) env;

    ReleaseAllClients(pCounter);
    FreeAllAlarms(pCounter);

    xfree(pCounter);
}

/*
*** Clean up an Await
 */
static void TriggerAwait(pWait)
WaitList *pWait;
{
    int n;
    ClientPtr client = pWait->aclient->client;
    int gone = client->clientGone;

    for (n=0; n < pWait->n_items; n++) {
	WaitItem *pItem = &pWait->items[n];
	SyncCounter *pCounter = pItem->counter;

	RemoveWaitItem(pCounter, pItem);
/*
 * Send an AdvanceNotify if necessary
 */
	if (!gone) {
	    CARD64 thresh;
	    C64Add(thresh, pItem->value, pItem->u.c_wait.threshold);

	    if (C64GreaterThan(pCounter->wait.value, thresh)) {
		xAdvanceNotifyEvent e;

		UpdateCurrentTimeIf();

		e.type = SyncEventBase + AdvanceNotify;
		e.sequenceNumber = client->sequence;
		e.counter = pCounter->id;
		e.wait_value_lo = C64Low32(pItem->value);
		e.wait_value_hi = C64High32(pItem->value);
		e.counter_value_lo = C64Low32(pCounter->wait.value);
		e.counter_value_hi = C64High32(pCounter->wait.value);
		e.time = currentTime.milliseconds;

		WriteEventsToClient(client, 1, (xEvent *)&e);
	    }
	}
    }

    pWait->aclient->wait = NULL;

    if (!gone) {
	AttendClient(client);
    }

    xfree(pWait);
}

/* ARGSUSED */
static void FreeAlarmClient(addr, id)
pointer addr;
XID id;
{
    AlarmClient *pClient = (AlarmClient *) addr;

    pClient->client = NULL;
    pClient->id = 0;
}

static void FreeAllAlarmClients(pAlarm)
SyncAlarm *pAlarm;
{
    AlarmClient *pClient;
    xAlarmNotifyEvent e;

    UpdateCurrentTimeIf();

    while ((pClient = pAlarm->clients) != 0) {
	ClientPtr client = pClient->client;

	if (client) {
	    if (!client->clientGone) {
		e.type = SyncEventBase + AlarmNotify;
		e.sequenceNumber = client->sequence;
		e.alarm = pClient->id;
		e.alarm_value_lo = C64Low32(pAlarm->wait.value);
		e.alarm_value_hi = C64High32(pAlarm->wait.value);
		e.counter_value_lo =
		    C64Low32(pAlarm->wait.counter->wait.value);
		e.counter_value_hi =
		    C64High32(pAlarm->wait.counter->wait.value);
		e.state = AlarmDestroyed;
		e.time = currentTime.milliseconds;
		WriteEventsToClient(client, 1, (xEvent *)&e);
	    }
	    FreeResource(pClient->id, RTAlarmClient);
	}

	pAlarm->clients = pClient->next;
	xfree(pClient);
    }
}

/* ARGSUSED */
static void FreeAlarm(addr, id)
pointer addr;
XID id;
{
    SyncAlarm *pAlarm = (SyncAlarm *) addr;
    SyncCounter *pCounter = pAlarm->wait.counter;

/*
 * Remove the wait for this alarm from the counter
 */
    if (pAlarm->state == AlarmActive) {
	WaitItem *pItem = pCounter->wait.next;

	for (;;) {
	    if (pItem==NULL) {
		fprintf(stderr, "Failed to find WaitItem\n");
		break;
	    }

	    if (pItem->is_alarm && pItem->u.a_wait.alarm==pAlarm) {
		RemoveWaitItem(pCounter, pItem);
		break;
	    }
	    pItem = pItem->next;
	}
    }

/*
 * Free all interested clients
 */
    FreeAllAlarmClients(pAlarm);

    xfree(pAlarm);
}

static void FreeAllAlarms(pCounter)
SyncCounter *pCounter;
{
    WaitItem *pItem;

/*
 * Look down the wait list for alarm waits on this alarm
 */
    for (pItem = pCounter->wait.next; pItem; pItem = pItem->next) {
	if (pItem->is_alarm) {
	    XID id = pItem->u.a_wait.alarm->id;

	    RemoveWaitItem(pCounter, pItem);
	    FreeResource(id, RT_NONE);

	    break;
	}
	pItem = pItem->next;
    }
}

/*
*** This client has died...
 */
/* ARGSUSED */
static void FreeAwaitClient(addr, id)
pointer addr;
XID id;
{
    AwaitClient *pClient = (AwaitClient *)addr;

    if (pClient->wait) {
	TriggerAwait(pClient->wait);
    }
    pClient->id = 0;
    pClient->client = NULL;
}

/*
*** Get info from the extension
 */
static int ProcSyncInfo(client)
ClientPtr client;
{
    REQUEST(xSyncInfoReq);
    xSyncInfoReply rep;
    int len, n;
    unsigned long *tmplist, *list;

    REQUEST_SIZE_MATCH(xSyncInfoReq);

    rep.type = X_Reply;
    rep.sequenceNumber = client->sequence;
    rep.majorVersion = SYNC_MAJOR_VERSION;
    rep.minorVersion = SYNC_MINOR_VERSION;
    rep.nCounters = NumSystemCounters;

    len = 0;
    for (n=0; n<NumSystemCounters; n++) {
	len += 1 + 2 + ((strlen(SysCounterList[n].name)>>2) + 1);
    }
    rep.length = len;

    tmplist = list = (unsigned long *)ALLOCATE_LOCAL(len<<2);
    if (!list) return BadAlloc;

    if (client->swapped) {
	swaps(&rep.sequenceNumber, n);
	swapl(&rep.length, n);
	swapl(&rep.nCounters, n);
    }

    for (n=0; n<NumSystemCounters; n++) {
	int l = strlen(SysCounterList[n].name);
	char *tmpname = (char *)tmplist;

	strncpy(tmpname, SysCounterList[n].name, l+1);
	tmplist += (l>>2) + 1;
	if (client->swapped) {
	    cpswapl(SysCounterList[n].counter->id, *tmplist);
	    tmplist++;
	    cpswapl(C64Low32(SysCounterList[n].resolution), *tmplist);
	    tmplist++;
	    cpswapl(C64High32(SysCounterList[n].resolution), *tmplist);
	    tmplist++;
	} else {
	    *tmplist++ = SysCounterList[n].counter->id;
	    *tmplist++ = C64Low32(SysCounterList[n].resolution);
	    *tmplist++ = C64High32(SysCounterList[n].resolution);
	}
    }

    WriteToClient(client, sizeof(xSyncInfoReply), (char *)&rep);
    WriteToClient(client, (len<<2), (char *)list);
    DEALLOCATE_LOCAL(list);

    return (client->noClientException);
}

static void DefaultQuery(pCounter, value_return)
SyncCounter *pCounter;
CARD64 *value_return;
{
    *value_return = pCounter->wait.value;
}

/*ARGSUSED*/
static void DefaultNewWait(pCounter, wait_value, wait_type, value_return)
SyncCounter *pCounter;
CARD64 wait_value;
int wait_type;
CARD64 *value_return;
{
    *value_return = pCounter->wait.value;
}

/*ARGSUSED*/
static void DefaultNoWait(pCounter)
SyncCounter *pCounter;
{
}

static void InitCounter(pCounter, initial)
SyncCounter *pCounter;
CARD64 initial;
{
    pCounter->wait.value = initial;
    pCounter->wait.next = NULL;
    pCounter->wait.prev = NULL;
    pCounter->wait.counter = pCounter;
    pCounter->owner = 0;
    pCounter->Query = DefaultQuery;
    pCounter->NewWait = DefaultNewWait;
    pCounter->NoWait = DefaultNoWait;
}

pointer SyncCreateSystemCounter(name, initial, resolution,
				query, newwait, nowait)
char *name;
CARD64 initial;
CARD64 resolution;
void (*query)();
void (*newwait)();
void (*nowait)();
{
    SyncCounter *pCounter;
    SysCounter *sysList;

    sysList = (SysCounter *)xrealloc(SysCounterList,
			sizeof(SysCounter)*(NumSystemCounters + 1));
    if (sysList == NULL) return NULL;

    pCounter = (SyncCounter *)xalloc(sizeof(SyncCounter));
    if (!pCounter) return NULL;

    if (RTCounter == 0) {
	RTCounter = CreateNewResourceType(FreeCounter);
	if (RTCounter == 0) {
	    return NULL;
	}
    }

    pCounter->id = FakeClientID(0);
    if (!AddResource(pCounter->id, RTCounter, (pointer)pCounter)) {
	xfree((pointer)pCounter);
	return NULL;
    }

    InitCounter(pCounter, initial);

    pCounter->Query = query;
    pCounter->NewWait = newwait;
    pCounter->NoWait = nowait;
    sysList[NumSystemCounters].name = name;
    sysList[NumSystemCounters].counter = pCounter;
    sysList[NumSystemCounters].resolution = resolution;

    SysCounterList = sysList;
    NumSystemCounters++;

    return (pointer) pCounter;
}

/*
*** Create a new counter
 */
static int ProcSyncCreateCounter(client)
ClientPtr client;
{
    REQUEST(xSyncCreateCounterReq);
    SyncCounter *pCounter;
    CARD64 initial;

    REQUEST_SIZE_MATCH(xSyncCreateCounterReq);

    LEGAL_NEW_RESOURCE(stuff->cid, client);

    pCounter = (SyncCounter *)xalloc(sizeof(SyncCounter));
    if (!pCounter) return BadAlloc;

    pCounter->id = stuff->cid;
    if (!AddResource(pCounter->id, RTCounter, (pointer)pCounter)) {
	xfree((pointer)pCounter);
	return BadAlloc;
    }

/*
 * Initialise the counter structure
 */
    cvtINT32sToC64(initial, stuff->initial_value_lo, stuff->initial_value_hi);
    InitCounter(pCounter, initial);
    pCounter->owner = client;

    return (client->noClientException);
}

/*
*** Destroy a counter
 */
static int ProcSyncDestroyCounter(client)
ClientPtr client;
{
    REQUEST(xSyncDestroyCounterReq);
    SyncCounter *pCounter;

    REQUEST_SIZE_MATCH(xSyncDestroyCounterReq);

    pCounter = (SyncCounter *)LookupIDByType(stuff->counter, RTCounter);
    if (pCounter==NULL || pCounter->owner==NULL) {
	client->errorValue = stuff->counter;
	return SyncErrorBase + BadCounter;
    }

    FreeResource(pCounter->id, RT_NONE);

    return Success;
}

/*
*** Advance a counter, waking clients as appropriate.
 */
void SyncAdvanceCounter(pc, amount)
pointer pc;
CARD64 amount;
{
    SyncCounter *pCounter = (SyncCounter *)pc;
    C64Add(pCounter->wait.value, pCounter->wait.value, amount);

    SyncReleaseClients(pCounter);
}

/*
*** Advance a counter to a new value, waking clients as appropriate.
 */
void SyncChangeCounter(pc, val)
pointer pc;
CARD64 val;
{
    SyncCounter *pCounter = (SyncCounter *)pc;
    pCounter->wait.value = val;

    SyncReleaseClients(pCounter);
}

/*
*** Await
 */
static int ProcSyncAwait(client)
ClientPtr client;
{
    REQUEST(xSyncAwaitReq);
    WaitList *pWait;
    int len, items;
    int n;
    XID *vlist;
    SyncCounter **counters;
    AwaitClient *pAClient;

    REQUEST_AT_LEAST_SIZE(xSyncAwaitReq);

    len = stuff->length - (sizeof(xSyncAwaitReq) >> 2);
/*
 * Each Wait is 6 words:
 */
    items = len/6;
    if (items==0 || items*6 != len) {
	return BadLength;
    }

    counters = (SyncCounter **)ALLOCATE_LOCAL(items*sizeof(SyncCounter *));
    pWait = (WaitList *)xalloc(sizeof(WaitList) +
			       sizeof(WaitItem)*(items-1));
    pWait->n_items = items;

/*
 * Check each item in turn for errors
 */
    vlist = (XID *)&stuff[1];

    for (n=0; n<items; n++) {
	WaitItem *pItem = &pWait->items[n];

	pItem->is_alarm = 0;
	counters[n] = pItem->counter =
	    (SyncCounter *)LookupIDByType(vlist[0], RTCounter);
	if (pItem->counter==NULL) {
	    client->errorValue = *vlist;
	    xfree(pWait);
	    DEALLOCATE_LOCAL(counters);
	    return SyncErrorBase + BadCounter;
	}

	pItem->value_type = vlist[1];
	pItem->u.c_wait.wait_ptr = pWait;
	switch (pItem->value_type) {
	case SyncAbsolute:
	case SyncRelative:
	    break;
	default:
	    client->errorValue = vlist[1];
	    xfree(pWait);
	    DEALLOCATE_LOCAL(counters);
	    return BadValue;
	}
	cvtINT32sToC64(pItem->value, vlist[2], vlist[3]);
	cvtINT32sToC64(pItem->u.c_wait.threshold, vlist[4], vlist[5]);

	vlist += 6;
    }

#ifdef XXX
/*
 * Add the resource for this await
 */
    pWait->id = FakeClientID(client->index);
    AddResource(pWait->id, RTAwait, (pointer)pWait);
#else /* XXX */
/*
 * Check that we know about this client already
 */
    for (pAClient = AwaitClientList; pAClient; pAClient = pAClient->next) {
	if (pAClient->client == client) {
	    break;
	}
    }
/*
 * If not, find an empty slot or add a new one
 */
    if (!pAClient) {
	for (pAClient = AwaitClientList; pAClient; pAClient = pAClient->next) {
	    if (pAClient->client == NULL) {
		break;
	    }
	}
	if (!pAClient) {
	    pAClient = (AwaitClient *)xalloc(sizeof(AwaitClient));
	    if (!pAClient) {
		DEALLOCATE_LOCAL(counters);
		return BadAlloc;
	    }
	    pAClient->next = AwaitClientList;
	    AwaitClientList = pAClient;
	}
/*
 * Add the resource for this client
 */
	pAClient->id = FakeClientID(client->index);
	pAClient->client = client;
	AddResource(pAClient->id, RTAwaitClient, (pointer)pAClient);
    }
    pAClient->wait = pWait;
    pWait->aclient = pAClient;
#endif /* XXX */

/*
 * Add each Item to the appropriate counter
 */
    for (n=0; n<items; n++) {
	WaitItem *pItem = &pWait->items[n];
	SyncCounter *pCounter = pItem->counter;

	AddWaitItem(pCounter, pItem);
    }

    IgnoreClient(client);

/*
 * Check each counter in case it was updated by the AddWaitItem
 * We use a local list of the counters because the WaitRec might be free-ed
 * within SyncReleaseClients.
 */
    for (n=0; n<items; n++) {
	SyncReleaseClients(counters[n]);
    }

    DEALLOCATE_LOCAL(counters);

    return Success;
}

static int ProcSyncAdvance(client)
ClientPtr client;
{
    REQUEST(xSyncAdvanceReq);
    SyncCounter *pCounter;
    CARD64 amount;

    REQUEST_SIZE_MATCH(xSyncAdvanceReq);

    pCounter = (SyncCounter *)LookupIDByType(stuff->counter, RTCounter);
    if (pCounter==NULL || pCounter->owner==NULL) {
	client->errorValue = stuff->counter;
	return SyncErrorBase + BadCounter;
    }

    cvtINT32sToC64(amount, stuff->amount_lo, stuff->amount_hi);
    SyncAdvanceCounter((pointer)pCounter, amount);

    return Success;
}

/*
*** Query a counter
 */
static int ProcSyncQueryCounter(client)
ClientPtr client;
{
    REQUEST(xSyncQueryCounterReq);
    xSyncQueryCounterReply rep;
    SyncCounter *pCounter;

    REQUEST_SIZE_MATCH(xSyncQueryCounterReq);

    pCounter = (SyncCounter *)LookupIDByType(stuff->counter, RTCounter);
    if (pCounter==NULL) {
	client->errorValue = stuff->counter;
	return SyncErrorBase + BadCounter;
    }

    rep.type = X_Reply;
    rep.length = 0;
    rep.sequenceNumber = client->sequence;

    if (pCounter->Query) (*pCounter->Query)((pointer)pCounter,
					    &pCounter->wait.value);

    rep.value_lo = C64Low32(pCounter->wait.value);
    rep.value_hi = C64High32(pCounter->wait.value);
    if (client->swapped) {
	register long n;
	swaps(&rep.sequenceNumber, n);
	swapl(&rep.length, n);
	swapl(&rep.value_lo, n);
	swapl(&rep.value_hi, n);
    }
    WriteToClient(client, sizeof(xSyncQueryCounterReply), (char *)&rep);
    return (client->noClientException);
}

/*
*** ParseAlarmAttributes
*** This is used by CreateAlarm and ChangeAlarm
 */
static int ParseAlarmAttributes(client, pAlarm, mask, values)
ClientPtr client;
SyncAlarm *pAlarm;
unsigned long mask;
XID *values;
{
    int events = 1;

    while (mask) {
	int index = lowbit (mask);

	mask &= ~index;
	switch (index) {
	case CACounter:
	    if (*values==0) {
		pAlarm->wait.counter = 0;
	    } else {
		pAlarm->wait.counter = (SyncCounter *)LookupIDByType(*values, RTCounter);
		if (!pAlarm->wait.counter) {
		    client->errorValue = *values;
		    return SyncErrorBase + BadCounter;
		}
	    }
	    values++;
	    break;
	case CAValue:
	    cvtINT32sToC64(pAlarm->wait.value, (INT32)values[0], (INT32)values[1]);
	    values += 2;
	    break;
	case CAIncrement:
	    cvtINT32sToC64(pAlarm->increment, (INT32)values[0], (INT32)values[1]);
	    values += 2;
	    break;
	case CALimit:
	    cvtINT32sToC64(pAlarm->limit, (INT32)values[0], (INT32)values[1]);
	    values += 2;
	    break;
	case CAValueType:
	    pAlarm->wait.value_type = (int)*values++;
	    break;
	case CALimitType:
	    pAlarm->limit_type = (int)*values++;
	    break;
	case CAEvents:
	    events = *values++;
	    break;
	default:
	    client->errorValue = mask;
	    return BadValue;
	}
    }

    if (events) {
	AlarmClient *pClient;

	for (pClient = pAlarm->clients; pClient;
	     pClient = pClient->next) {
	    if (pClient->client == client) break;
	}
	if (!pClient) { /* Really Enable */
	    for (pClient = pAlarm->clients; pClient;
		 pClient = pClient->next) {
		if (pClient->client == 0) break;
	    }
	    if (!pClient) {
		pClient = (AlarmClient *)xalloc(sizeof(AlarmClient));
		if (!pClient) return BadAlloc;
		pClient->next = pAlarm->clients;
		pAlarm->clients = pClient;
	    }
	    pClient->client = client;
	    pClient->id = FakeClientID(client->index);
	    if (!AddResource(pClient->id, RTAlarmClient, pClient)) {
		return BadAlloc;
	    }
	}
    } else {
	AlarmClient *pClient;

	for (pClient = pAlarm->clients; pClient;
	     pClient = pClient->next) {
	    if (pClient->client == client) {
		FreeResource(pClient->id, RT_NONE);
		break;
	    }
	}
    }

    return Success;
}

/*
*** Create Alarm
 */
static int ProcSyncCreateAlarm(client)
ClientPtr client;
{
    REQUEST(xSyncCreateAlarmReq);
    SyncAlarm *pAlarm;
    int status;
    int len, vmask;
    XID *vlist;

    REQUEST_AT_LEAST_SIZE(xSyncCreateAlarmReq);

    LEGAL_NEW_RESOURCE(stuff->id, client);

    vmask = stuff->valueMask;
    len = stuff->length - (sizeof(xSyncCreateAlarmReq) >> 2);
/*
 * Check the length
 * The Value, Increment and Limit fields need 2 words each
 */
    if (len != Ones(vmask) + Ones(vmask & (CAValue | CAIncrement | CALimit))) {
	return BadLength;
    }

    vlist = (XID *)&stuff[1];
 
    if (!(pAlarm = (SyncAlarm *)xalloc(sizeof(SyncAlarm)))) {
	return BadAlloc;
    }

    pAlarm->id = stuff->id;
    pAlarm->client = client;
    pAlarm->clients = NULL;
    pAlarm->state = AlarmActive;
    cvtLongToC64(pAlarm->increment, 1L);
    pAlarm->limit_type = SyncAbsolute;
    cvtLongToC64(pAlarm->limit, 0L);
    pAlarm->wait.counter = 0;
    pAlarm->wait.value_type = SyncAbsolute;
    cvtLongToC64(pAlarm->wait.value, 0L);
    pAlarm->wait.next = 0;
    pAlarm->wait.prev = 0;
    pAlarm->wait.is_alarm = 1;
    pAlarm->wait.u.a_wait.alarm = pAlarm;

    status = ParseAlarmAttributes(client, pAlarm, vmask, vlist);

    if (status != Success) {
	xfree(pAlarm);
	return status;
    }

    if (!AddResource(stuff->id, RTAlarm, pAlarm)) {
	xfree(pAlarm);
	return BadAlloc;
    }

    if (pAlarm->wait.counter) {
	AddWaitItem(pAlarm->wait.counter, &pAlarm->wait);

	if (pAlarm->limit_type==SyncRelative) {
	    C64Add(pAlarm->limit, pAlarm->wait.counter->wait.value, pAlarm->limit);
	    pAlarm->limit_type=SyncAbsolute;
	}

	TriggerAlarm(pAlarm);
	SyncReleaseClients(pAlarm->wait.counter);
    } else {
	pAlarm->state = AlarmExhausted;
    }

    return Success;
}

/*
*** Change Alarm
 */
static int ProcSyncChangeAlarm(client)
ClientPtr client;
{
    REQUEST(xSyncChangeAlarmReq);
    SyncAlarm *pAlarm;
    int vmask;
    XID *vlist;
    int len, status;
    SyncCounter *pOldCounter;

    REQUEST_AT_LEAST_SIZE(xSyncChangeAlarmReq);

    pAlarm = (SyncAlarm *)LookupIDByType(stuff->alarm, RTAlarm);
    if (!pAlarm) {
	client->errorValue = stuff->alarm;
	return SyncErrorBase + BadAlarm;
    }

    vmask = stuff->valueMask;
    len = stuff->length - (sizeof(xSyncChangeAlarmReq) >> 2);
    if (len != Ones(vmask) + Ones(vmask & (CAValue | CAIncrement | CALimit))) {
	return BadLength;
    }

    vmask = stuff->valueMask;
    vlist = (XID *)&stuff[1];

    pOldCounter = pAlarm->wait.counter;
    status = ParseAlarmAttributes(client, pAlarm, vmask, vlist);

    if (status != Success) {
	return status;
    }

    if (pOldCounter != pAlarm->wait.counter) {
	RemoveWaitItem(pOldCounter, &pAlarm->wait);
	AddWaitItem(pAlarm->wait.counter, &pAlarm->wait);
    }
    if (pAlarm->limit_type==SyncRelative) {
	C64Add(pAlarm->limit, pAlarm->wait.counter->wait.value, pAlarm->limit);
	pAlarm->limit_type=SyncAbsolute;
    }

    pAlarm->state = AlarmActive;
    TriggerAlarm(pAlarm);
    SyncReleaseClients(pAlarm->wait.counter);

    return Success;
}

static int ProcSyncQueryAlarm(client)
ClientPtr client;
{
    REQUEST(xSyncQueryAlarmReq);
    SyncAlarm *pAlarm;
    xSyncQueryAlarmReply rep;

    REQUEST_SIZE_MATCH(xSyncQueryAlarmReq);
  
    pAlarm = (SyncAlarm *)LookupIDByType(stuff->alarm, RTAlarm);
    if (!pAlarm) {
	client->errorValue = stuff->alarm;
	return SyncErrorBase + BadAlarm;
    }

    rep.type = X_Reply;
    rep.length = 0;
    rep.sequenceNumber = client->sequence;
  
    rep.counter = pAlarm->wait.counter->id;
    rep.value_lo = C64Low32(pAlarm->wait.value);
    rep.value_hi = C64High32(pAlarm->wait.value);
    rep.increment_lo = C64Low32(pAlarm->increment);
    rep.increment_hi = C64High32(pAlarm->increment);
    rep.limit_lo = C64Low32(pAlarm->limit);
    rep.limit_hi = C64High32(pAlarm->limit);

    if (client->swapped) {
	register long n;
	swaps(&rep.sequenceNumber, n);
	swapl(&rep.length, n);
	swapl(&rep.counter, n);
	swapl(&rep.value_lo, n);
	swapl(&rep.value_hi, n);
	swapl(&rep.increment_lo, n);
	swapl(&rep.increment_hi, n);
	swapl(&rep.limit_lo, n);
	swapl(&rep.limit_hi, n);
    }
    WriteToClient(client, sizeof(xSyncQueryAlarmReply), (char *)&rep);

    return (client->noClientException);
}


static int ProcSyncDestroyAlarm(client)
ClientPtr client;
{
    REQUEST(xSyncDestroyAlarmReq);
  
    REQUEST_SIZE_MATCH(xSyncDestroyAlarmReq);
  
    if (!LookupIDByType(stuff->alarm, RTAlarm)) {
	client->errorValue = stuff->alarm;
	return SyncErrorBase + BadAlarm;
    }

    FreeResource(stuff->alarm, RT_NONE);      

    return (client->noClientException);
}

/*
*** Given an extension request, call the appropriate request procedure
 */
static int ProcSyncDispatch(client)
ClientPtr client;
{
    REQUEST(xReq);
    switch (stuff->data) {

    case X_SyncInfo:
	return ProcSyncInfo(client);
    case X_SyncCreateCounter:
	return ProcSyncCreateCounter(client);
    case X_SyncDestroyCounter:
	return ProcSyncDestroyCounter(client);
    case X_SyncAdvance:
	return ProcSyncAdvance(client);
    case X_SyncAwait:
	return ProcSyncAwait(client);
    case X_SyncQueryCounter:
	return ProcSyncQueryCounter(client);
    case X_SyncCreateAlarm:
	return ProcSyncCreateAlarm(client);
    case X_SyncChangeAlarm:
	return ProcSyncChangeAlarm(client);
    case X_SyncQueryAlarm:
	return ProcSyncQueryAlarm(client);
    case X_SyncDestroyAlarm:
	return ProcSyncDestroyAlarm(client);
    default:
	return BadRequest;
    }
}

/*
 * Boring Swapping stuff ...
 */

static int SProcSyncInfo(client)
ClientPtr client;
{
    REQUEST(xSyncInfoReq);
    register char n;

    swaps(&stuff->length, n);

    return ProcSyncInfo(client);
}

static int SProcSyncCreateCounter(client)
ClientPtr client;
{
    REQUEST(xSyncCreateCounterReq);
    register char n;

    swaps(&stuff->length, n);
    swapl(&stuff->cid, n);
    swapl(&stuff->initial_value_lo, n);
    swapl(&stuff->initial_value_hi, n);

    return ProcSyncCreateCounter(client);
}

static int SProcSyncDestroyCounter(client)
ClientPtr client;
{
    REQUEST(xSyncDestroyCounterReq);
    register char n;

    swaps(&stuff->length, n);
    swapl(&stuff->counter, n);

    return ProcSyncDestroyCounter(client);
}

static int SProcSyncAwait(client)
ClientPtr client;
{
    REQUEST(xSyncAwaitReq);
    register char n;

    swaps(&stuff->length, n);
    SwapRestL(stuff);

    return ProcSyncAwait(client);
}

static int SProcSyncAdvance(client)
ClientPtr client;
{
    REQUEST(xSyncAdvanceReq);
    register char n;

    swaps(&stuff->length, n);
    swapl(&stuff->counter, n);
    swapl(&stuff->amount_lo, n);
    swapl(&stuff->amount_hi, n);

    return ProcSyncAdvance(client);
}

static int SProcSyncQueryCounter(client)
ClientPtr client;
{
    REQUEST(xSyncQueryCounterReq);
    register char n;

    swaps(&stuff->length, n);
    swapl(&stuff->counter, n);

    return ProcSyncQueryCounter(client);
}

static int SProcSyncCreateAlarm(client)
ClientPtr client;
{
    REQUEST(xSyncCreateAlarmReq);
    register char n;

    swaps(&stuff->length, n);
    swapl(&stuff->id, n);
    swapl(&stuff->valueMask, n);
    SwapRestL(stuff);

    return ProcSyncCreateAlarm(client);
}

static int SProcSyncChangeAlarm(client)
ClientPtr client;
{
    REQUEST(xSyncChangeAlarmReq);
    register char n;

    REQUEST_AT_LEAST_SIZE(xSyncChangeAlarmReq);
    swaps(&stuff->length, n);
    swapl(&stuff->alarm, n);
    swapl(&stuff->valueMask, n);
    SwapRestL(stuff);

    return ProcSyncChangeAlarm(client);
}

static int SProcSyncQueryAlarm(client)
ClientPtr client;
{
    REQUEST(xSyncQueryAlarmReq);
    register char n;

    swaps(&stuff->length, n);
    swapl(&stuff->alarm, n);

    return ProcSyncQueryAlarm(client);
}

static int SProcSyncDestroyAlarm(client)
ClientPtr client;
{
    REQUEST(xSyncDestroyAlarmReq);
    register char n;

    swaps(&stuff->length, n);
    swapl(&stuff->alarm, n);

    return ProcSyncDestroyAlarm(client);
}

static int SProcSyncDispatch(client)
ClientPtr client;
{
    REQUEST(xReq);
    switch (stuff->data) {

    case X_SyncInfo:
	return SProcSyncInfo(client);
    case X_SyncCreateCounter:
	return SProcSyncCreateCounter(client);
    case X_SyncDestroyCounter:
	return SProcSyncDestroyCounter(client);
    case X_SyncAdvance:
	return SProcSyncAdvance(client);
    case X_SyncAwait:
	return SProcSyncAwait(client);
    case X_SyncQueryCounter:
	return SProcSyncQueryCounter(client);
    case X_SyncCreateAlarm:
	return SProcSyncCreateAlarm(client);
    case X_SyncChangeAlarm:
	return SProcSyncChangeAlarm(client);
    case X_SyncQueryAlarm:
	return SProcSyncQueryAlarm(client);
    case X_SyncDestroyAlarm:
	return SProcSyncDestroyAlarm(client);
    default:
	return BadRequest;
    }
}

/*
 * Event Swapping
 */

static void SAdvanceNotifyEvent(from, to)
xAdvanceNotifyEvent *from, *to;
{
    to->type = from->type;
    to->kind = from->kind;
    cpswaps(from->sequenceNumber, to->sequenceNumber);
    cpswapl(from->counter, to->counter);
    cpswapl(from->wait_value_lo, to->wait_value_lo);
    cpswapl(from->wait_value_hi, to->wait_value_hi);
    cpswapl(from->counter_value_lo, to->counter_value_lo);
    cpswapl(from->counter_value_hi, to->counter_value_hi);
    cpswapl(from->time, to->time);
}

static void SAlarmNotifyEvent(from, to)
xAlarmNotifyEvent *from, *to;
{
    to->type = from->type;
    cpswaps(from->sequenceNumber, to->sequenceNumber);
    cpswapl(from->alarm, to->alarm);
    cpswapl(from->alarm_value_lo, to->alarm_value_lo);
    cpswapl(from->alarm_value_hi, to->alarm_value_hi);
    cpswapl(from->counter_value_lo, to->counter_value_lo);
    cpswapl(from->counter_value_hi, to->counter_value_hi);
    cpswapl(from->time, to->time);
    to->state = from->state;
}

/*
*** Close everything down.
*** This is fairly simple for now.
 */
/*ARGSUSED*/
static void SyncResetProc(extEntry)
ExtensionEntry	*extEntry;
{
    xfree(SysCounterList);
    SysCounterList = NULL;
    NumSystemCounters = 0;
    RTCounter = 0;
}

/*
*** Initialise the extension.
 */
void SyncExtensionInit()
{
    ExtensionEntry *extEntry;

    if (RTCounter == 0) {
	RTCounter = CreateNewResourceType(FreeCounter);
    }
    RTAwaitClient = CreateNewResourceType(FreeAwaitClient);
    RTAlarm = CreateNewResourceType(FreeAlarm);
    RTAlarmClient = CreateNewResourceType(FreeAlarmClient);

    if (RTCounter == NULL ||
	RTAwaitClient == NULL ||
	RTAlarm == NULL ||
	RTAlarmClient == NULL ||
	(extEntry = AddExtension(SYNCNAME,
				 SyncNumberEvents, SyncNumberErrors,
				 ProcSyncDispatch, SProcSyncDispatch,
				 SyncResetProc,
				 StandardMinorOpcode))==NULL) {
	fprintf(stderr, "Sync Extension %d.%d failed to Initialise\n",
		SYNC_MAJOR_VERSION, SYNC_MINOR_VERSION);
	return;
    }

    SyncReqCode = extEntry->base;
    SyncEventBase = extEntry->eventBase;
    SyncErrorBase = extEntry->errorBase;
    EventSwapVector[SyncEventBase + AdvanceNotify] = SAdvanceNotifyEvent;
    EventSwapVector[SyncEventBase + AlarmNotify] = SAlarmNotifyEvent;

/*
 * Although SERVERTIME is implemented by the OS layer, we initialise it here
 * because doing it in OsInit() is too early. The resource database is not
 * initialised when OsInit() is called.
 * This is just about OK because there is always a servertime counter.
 */
    SyncInitServertime();

#ifdef DEBUG
    fprintf(stderr, "Sync Extension %d.%d\n",
	    SYNC_MAJOR_VERSION, SYNC_MINOR_VERSION);
#endif
}
#endif /* SYNC */
